@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2009 Sam Trenholme
+/* Copyright (c) 2007-2010 Sam Trenholme
  *
  * TERMS
  *
@@ -243,7 +243,7 @@ void dwm_init_fsm() {
 
 /* Fatal error while parsing Mararc file */
 void dwm_fatal(char *why) {
-        dw_alog_number("Fatal error in dwood2rc file on line ",dwm_linenum,
+        dw_alog_number("Fatal error in dwood3rc file on line ",dwm_linenum,
                 why);
         exit(1);
 }
@@ -368,7 +368,7 @@ void dwm_init_mararc() {
         dwm_dict_init();
         /* Numeric mararc variables have default values.  */
         key_n[DWM_N_maxprocs] = 32;
-        key_n[DWM_N_timeout_seconds] = 3;
+        key_n[DWM_N_timeout_seconds] = 2;
         key_n[DWM_N_dns_port] = 53;
         key_n[DWM_N_upstream_port] = 53;
         key_n[DWM_N_handle_overload] = 1;
@@ -389,6 +389,10 @@ void dwm_init_mararc() {
         key_n[DWM_N_ttl_age] = 1;
         key_n[DWM_N_max_inflights] = 8;
         key_n[DWM_N_deliver_all] = 1;
+        key_n[DWM_N_filter_rfc1918] = 1;
+        key_n[DWM_N_ns_glueless_type] = 1;
+        key_n[DWM_N_reject_aaaa] = 0;
+        key_n[DWM_N_reject_mx] = 1;
 }
 
 /* Look for a Mararc parameter; -1 if not found/error; 0-n if found
@@ -499,9 +503,13 @@ void dwm_execfile(dw_str *execfile, dw_str *fname) {
  */
 
 int dwm_set_todo(dw_str **actions) {
-        if(actions[1] == 0 || actions[6] != 0) { /* Do nothing */
+        if(actions[1] == 0) { /* Do nothing */
                 return 0;
         }
+        if(actions[6] != 0 && actions[1] != 0 && actions[2] == 0) {
+                /* Init dictionary variable */
+                return 6;
+        }
         if(actions[7] != 0) { /* Read and parse other file */
                 dwm_execfile(actions[1],actions[7]);
                 return 0;
@@ -543,17 +551,41 @@ int dwm_set_todo(dw_str **actions) {
  *
  * Global variables modified: key_d
  */
-void dwm_dict_add(int num, dw_str *key, dw_str *value) {
-        if(num >= KEY_D_COUNT) { /* Sanity check */
+void dwm_dict_add(int num, dw_str *key, dw_str *value, int todo) {
+        dw_str *temp = 0;
+
+        if(num >= KEY_D_COUNT || num < 0) { /* Sanity check */
+                dwm_fatal("Unknown dictionary variable");
+        }
+
+        if(todo == 6 && (key != 0 || value != 0)) {
+                dwm_fatal("Illegal dictionary initialization");
+        }
+
+        /* Initialize dictionary if needed*/
+        if(key_d[num] == 0 && todo == 6) { /* Initialized with {} */
+                key_d[num] = dwd_init();
+                return;
+        } else if(key_d[num] == 0) {
+                dwm_fatal("Uninitialized dictionary variable");
+        }
+
+        if(todo != 4) { /* =, create new dictionary element */
+                key_d[num] = dwd_add(key_d[num],key,value);
                 return;
         }
 
-        /* Initialize element if needed*/
-        if(key_d[num] == 0) {
-                key_d[num] = dwd_init();
+        /* OK, += so we append an already existing element */
+        temp = dwd_fetch(key_d[num],key);
+        if(temp == 0) {
+                dwm_fatal("Appending unset dictionary index");
         }
 
-        key_d[num] = dwd_add(key_d[num],key,value);
+        dw_append(value,temp);
+
+        dwd_add(key_d[num],key,temp);
+
+        dw_destroy(temp);
 
 }
 
@@ -594,9 +626,12 @@ void dwm_set_keys(dw_str **actions) {
         if(todo == 1 || todo == 2) { /* Normal variable */
                 num = dwm_grep_params(actions[1],key_s_names,KEY_S_COUNT);
                 list = key_s;
-        } else if(todo == 3 || todo == 4) { /* Dictionary variable */
+        } else if(todo == 3 || todo == 4 || todo == 6) { /* Dict. variable */
                 num = dwm_grep_params(actions[1],key_d_names,KEY_D_COUNT);
-                dwm_dict_add(num, actions[2], actions[3]);
+                if(num < 0) { /* Make sure the parameter is a legal one */
+                        dwm_fatal("Unknown dwood3rc dictionary parameter");
+                }
+                dwm_dict_add(num, actions[2], actions[3], todo);
                 return;
         } else if(todo == 5) { /* Numeric variable */
                 int32_t val;
@@ -606,7 +641,7 @@ void dwm_set_keys(dw_str **actions) {
                         dwm_fatal("Invalid numeric value");
                 }
                 if(num < 0) { /* Make sure the parameter is a legal one */
-                        dwm_fatal("Unknown deadwoodrc parameter");
+                        dwm_fatal("Unknown dwood3rc numeric parameter");
                 }
                 key_n[num] = val;
                 return;
@@ -615,14 +650,14 @@ void dwm_set_keys(dw_str **actions) {
         }
 
         if(list == 0) { /* Sanity check */
-                dwm_fatal("Unknown deadwoodrc action");
+                dwm_fatal("Unknown dwood3rc action");
         }
         if(num < 0) { /* Make sure the parameter is a legal one */
-                dwm_fatal("Unknown deadwoodrc parameter");
+                dwm_fatal("Unknown dwood3rc string parameter");
         }
 
         if(list[num] == 0 && (todo & 1) == 0) {
-                dwm_fatal("Appending to unset parameter");
+                dwm_fatal("Appending to unset string parameter");
         } else if(list[num] != 0 && (todo & 1) == 0 /* append */) {
                 dw_append(actions[3],list[num]);
                 return;
