@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2009 Sam Trenholme and Jean-Jacques Sarton
+/* Copyright (c) 2007-2010 Sam Trenholme and Jean-Jacques Sarton
  *
  * TERMS
  *
@@ -19,6 +19,7 @@
 #include "DwSocket.h"
 #include "DwTcpSocket.h"
 #include "DwSys.h"
+#include "DwRecurse.h"
 
 /* Mararc parameters that are set in DwMararc.c */
 extern dw_str *key_s[];
@@ -36,7 +37,6 @@ extern SOCKET tcp_b_local[]; /* Local TCP sockets */
 /* List of addresses we will bind to */
 ip_addr_T bind_address[DW_MAXIPS + 1];
 ip_addr_T upstream_address[DW_MAXIPS + 1];
-dwd_dict *upstream_dict;
 
 /* List of who is allowed to use Deadwood */
 ip_mask_T recursive_acl[DW_MAXIPS + 1];
@@ -51,7 +51,7 @@ remote_T *rem;
 /* The cache */
 extern dw_hash *cache;
 
-/* Some dwood2rc parameters */
+/* Some dwood3rc parameters */
 int maxprocs = 32;
 int max_tcp_procs = 8;
 int timeout_seconds = 3;
@@ -67,7 +67,6 @@ int32_t maradns_uid = 99;
 int32_t maradns_gid = 99;
 int num_retries = 2;
 dwd_dict *blacklist_dict = 0;
-int deliver_all = 0;
 
 #ifdef MINGW
 u_long dont_block = 0;
@@ -293,6 +292,8 @@ void pop_ip_mask(dw_str *list, ip_mask_T *ip_mask) {
 
         q = strchr(c,'/');
         if(q == 0 || *q != '/') {
+                pop_ip_core(list,(ip_addr_T *)ip_mask,c);
+                make_netmask(ip_mask->len * 8,ip_mask->mask,ip_mask->len);
                 goto catch_pop_ip_mask;
         }
         *q = 0; /* pop_ip_core can't take an IP ending with a slash */
@@ -406,6 +407,9 @@ int check_ip_acl(ip_addr_T *ip) {
         uint8_t m;
 
         for(a = 0; a < DW_MAXIPS; a++) {
+                if(recursive_acl[a].len == 0) {
+                        return 0;
+                }
                 if(ip->len != recursive_acl[a].len) {
                         continue;
                 }
@@ -430,7 +434,7 @@ int check_ip_acl(ip_addr_T *ip) {
  * in dwm_init_mararc() in the file DwMararc.c */
 void process_numeric_mararc_params() {
 
-        maxprocs =        get_key_n(DWM_N_maxprocs,       8,1024,-1);
+        maxprocs =        get_key_n(DWM_N_maxprocs,       8,16384,-1);
         max_tcp_procs =   get_key_n(DWM_N_max_tcp_procs,  4,1024,-1);
         timeout_seconds = get_key_n(DWM_N_timeout_seconds,1,300,-1);
         timeout_seconds_tcp = get_key_n(DWM_N_timeout_seconds_tcp,1,300,-1);
@@ -444,7 +448,6 @@ void process_numeric_mararc_params() {
         maradns_gid = get_key_n(DWM_N_maradns_gid,10,65535,99);
         resurrections = get_key_n(DWM_N_resurrections,0,1,1);
         num_retries = get_key_n(DWM_N_num_retries,0,8,2);
-        deliver_all = get_key_n(DWM_N_deliver_all,0,1,1);
 
         if((num_ports & (num_ports - 1)) != 0) {
                 dw_fatal("num_ports must be a power of 2");
@@ -454,45 +457,67 @@ void process_numeric_mararc_params() {
 }
 
 /* Read and handle the upstream_servers dictionary variable */
-void process_upstream_servers() {
-        dw_str *q = 0 , *upstream = 0, *r = 0, *s = 0;
-        int a = 0;
+int process_root_upstream_servers(int param, int is_upstream, char *bad) {
+        dw_str *q = 0, *r = 0, *s = 0, *ns_refer = 0;
+        int a = 0, out = 0;
 
-        q = dw_create(32);
-        dw_qrappend((uint8_t *)".",q,0);
-        upstream = dwm_dict_fetch(DWM_D_upstream_servers,q);
-        dw_destroy(q);
-        set_ip_list(upstream_address,upstream);
-        if(upstream == 0) {
-                dw_fatal("Could not get upstream_servers");
-        }
-        dw_destroy(upstream);
-
-        upstream_dict = dwd_init();
         q = 0;
-        /* Fill up the upstream_dict hash */
+        /* Add upstream server elements to the main cache */
         for(a=0;a<20000;a++) {
-                r = dwm_dict_nextkey(DWM_D_upstream_servers,q);
+                r = dwm_dict_nextkey(param,q);
                 dw_destroy(q);
                 if(r == 0) {
                         break;
                 }
-                s = dwm_dict_fetch(DWM_D_upstream_servers,r);
+                out = 1;
+                s = dwm_dict_fetch(param,r);
                 q = dw_dnsname_convert(r);
+                ns_refer = dwx_ns_convert(s,is_upstream,q);
                 if(q == 0) {
-                        dw_log_dwstr_str(
-                                "Bad line in configuration file looks like: "
-                                "upstream_servers[\"",r,"\"]",0);
+                        dw_log_dwstr_str(bad,r,"\"]",0);
                         dw_fatal("Please fix this line");
                 }
-                dw_put_u16(q, 65392, -1); /* Find upstream private RR */
-                upstream_dict = dwd_add(upstream_dict,q,s);
+                dw_put_u16(q, 65395, -1); /* NS refer private RR */
+                dwh_add(cache,q,ns_refer,1,2);
                 dw_destroy(q);
                 dw_destroy(s);
+                dw_destroy(ns_refer);
                 q = dw_copy(r);
                 dw_destroy(r);
         }
 
+        return out;
+
+}
+
+/* Process the root_servers and upstream_servers values, using
+ * default root servers if needed
+ */
+void process_root_upstream() {
+        int elements_in_cache = 0;
+        dw_str *s = 0, *q = 0, *ns_refer = 0;
+
+        elements_in_cache +=
+                process_root_upstream_servers(DWM_D_upstream_servers,1,
+                        "Bad dwood3rc line looks like:\nupstream_servers[\"");
+        elements_in_cache +=
+                process_root_upstream_servers(DWM_D_root_servers,0,
+                        "Bad dwood3rc line looks like:\nroot_servers[\"");
+
+        if(elements_in_cache == 0) { /* Use default root servers */
+                dw_log_3strings("Using default ICANN root servers:"," ",
+                        ROOT_SERVERS,1);
+                s = dw_create(256);
+                q = dw_create(16);
+                dw_qrappend((uint8_t *)ROOT_SERVERS,s,0);
+                ns_refer = dwx_ns_convert(s,0,0);
+                dw_put_u8(q,0,-1); /* Root server (".") */
+                dw_put_u16(q,65395,-1); /* NS refer private RR */
+                dwh_add(cache,q,ns_refer,1,2);
+                dw_destroy(q);
+                dw_destroy(s);
+                dw_destroy(ns_refer);
+        }
 }
 
 /* Read mararc parameters and set global variables based on those
@@ -500,8 +525,6 @@ void process_upstream_servers() {
 void process_mararc_params() {
         dw_str *bind = 0, *r_acl = 0;
 
-        process_upstream_servers();
-
         bind = get_bind_addrs();
         if(bind == 0) {
                 dw_fatal("Could not bind");
@@ -545,9 +568,12 @@ void process_mararc_params() {
 void init_b_remote() {
         int a = 0;
         for(a = 0; a < maxprocs; a++) {
-                b_remote[a] = -1;
+                b_remote[a] = INVALID_SOCKET;
+                rem[a].socket = INVALID_SOCKET;
                 rem[a].query = 0;
+                rem[a].glueless = 0;
                 rem[a].local = 0;
+                rem[a].ns = 0;
                 reset_rem(a);
         }
 }
@@ -624,17 +650,16 @@ void set_rx_fd(fd_set *rx_fd) {
 int32_t find_free_remote() {
         int32_t a = 0;
         for(a = 0; a < maxprocs; a++) {
-                if(rem[a].socket == -1) { /* Available for use */
+                if(rem[a].socket == INVALID_SOCKET &&
+                                rem[a].local == 0) { /* Available for use */
                         return a;
                 }
         }
         return -1; /* None are available (We're overloaded) */
 }
 
-/* Get an upstream IP address */
-ip_addr_T get_upstream_ip(dw_str *query) {
-        ip_addr_T addr = {0,{0,0}};
-        ip_addr_T upstream_addr[DW_MAXIPS + 1];
+/* Get an upstream NS delegation */
+dw_str *get_upstream_ns(dw_str *query, int connection_number) {
         dw_str *q = 0, *r = 0;
         int a = 0;
 
@@ -642,11 +667,11 @@ ip_addr_T get_upstream_ip(dw_str *query) {
          * of upstream servers */
         r = dw_copy(query);
         dw_pop_u16(r);
-        dw_put_u16(r, 65392, -1); /* Special RR type for "upstream server" */
+        dw_put_u16(r, 65395, -1); /* Special "NS refer" RR type */
 
         /* Figure out which upstream server to use */
         for(a=0; q == 0 && a < 260; a++) {
-                q = dwd_fetch(upstream_dict,r);
+                q = dwh_get(cache,r,0,1);
                 if(q == 0) {
                         q = dw_dnslabel_chop(r);
                         dw_destroy(r);
@@ -655,33 +680,39 @@ ip_addr_T get_upstream_ip(dw_str *query) {
                         dw_destroy(q);
                         q = 0;
                 }
+                if(r == 0) {
+                        break;
+                }
         }
 
         dw_destroy(r);
         r = 0;
-        /* Convert the upstream_server ASCII list of IPs in to binary IPs */
-        set_ip_list(upstream_addr,q);
-        dw_destroy(q);
-        q = 0;
 
-        /* Choose an upstream server at random */
-        for(a=0; a < DW_MAXIPS; a++) {
-                if(upstream_addr[a].len == 0) {
-                        break;
-                }
-        }
-        if(a != 0) {
-                a = dwr_rng(rng_seed) % a;
-        } else {
-                return addr;
-        }
-        if(upstream_addr[a].len == 0) { /* Sanity check */
-                return addr;
+        /* Is this an upstream referral? */
+        if(dw_fetch_u8(q,-1) == TYPE_UPSTREAM_REFER) {
+                rem[connection_number].is_upstream = 1;
         }
 
-        /* And return the upstream server we have chosen */
-        addr = upstream_addr[a];
-        return addr;
+        /* Return the list of NS servers */
+        return q;
+}
+
+/* Given a UDP remote connection, decide on an IP to send a query to
+ * (or, if needed, a gluless NS record); if the list of nameservers
+ * to contact is empty, fill it up using get_upstream_ns */
+ip_addr_T get_upstream_ip(dw_str *query, int b) {
+        ip_addr_T addr = {0,{0,0},0,0};
+
+        if(rem[b].ns == 0) {
+                rem[b].ns = get_upstream_ns(query,b);
+        }
+
+        if(rem[b].ns == 0) {
+                return addr; /* Error; returns invalid addr */
+        }
+
+        return dwx_ns_getip(rem[b].ns,rng_seed,b);
+
 }
 
 /* This bit of code is used five times in process_results() to see if
@@ -734,6 +765,9 @@ int handle_resurrections(int a) {
         }
 
         for(local_num = 0; local_num < rem[a].num_locals; local_num++) {
+                if(rem[a].local[local_num]->glueless_type != 0) {
+                        continue;
+                }
                 packet = make_dns_packet(rem[a].query,value,
                                 rem[a].local[local_num]->local_id);
                 if(packet == 0) {
@@ -747,6 +781,10 @@ int handle_resurrections(int a) {
                         tcp_return_reply(rem[a].local[local_num]->tcp_num,
                                 (void *)packet->str, packet->len);
                 }
+                if(packet != 0) {
+                        dw_destroy(packet);
+                        packet = 0;
+                }
         }
 
         ret = 1;
@@ -778,17 +816,23 @@ int handle_expired(int a) {
                 closesocket(rem[a].socket);
                 rem[a].socket = INVALID_SOCKET;
                 /* 0x0180: QR = 0; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
-                 *         Z = 0; RCODE = 0 */
+                 *         Z = 0; RCODE = 0; 0x080 is same but with RD = 0 */
                 rem[a].remote_id = dwr_rng(rng_seed);
-                packet = make_dns_header(rem[a].remote_id,0x0180,0,0,0);
+                if(rem[a].is_upstream == 1) {
+                        packet = make_dns_header(rem[a].remote_id,0x180,0,0,0);
+                } else {
+                        packet = make_dns_header(rem[a].remote_id,0x080,0,0,0);
+                }
                 if(dw_append(rem[a].query,packet) == -1 ||
                    dw_put_u16(packet,1,-1) == -1 /* QCLASS: 1 */) {
                         goto catch_handle_expired;
                 }
+                dw_log_dwstr_str("Connection for query ",rem[a].query,
+                                " did not respond; trying again",128);
                 make_remote_connection(a,(unsigned char *)packet->str,
-                                       packet->len,rem[a].query);
+                                packet->len,rem[a].query,INVALID_SOCKET);
                 rem[a].die = get_time() + ((int64_t)timeout_seconds << 8);
-                ret = 1;
+                ret = 1; /* Do not kill; we're trying again */
                 goto catch_handle_expired;
         }
         if(resurrections == 1) {
@@ -824,7 +868,7 @@ void kill_expired() {
                                 continue;
                         }
                         closesocket(rem[a].socket);
-                        b_remote[a] = -1;
+                        b_remote[a] = INVALID_SOCKET;
                         reset_rem(a);
                 }
         }
