@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2009 Sam Trenholme
+/* Copyright (c) 2007-2010 Sam Trenholme
  *
  * TERMS
  *
@@ -126,6 +126,80 @@ void dw_log_dwstr_p(char *s1, dw_str *s2, int min_log_level) {
         }
 }
 
+/* Log an IP; private */
+void dw_log_ip_p(ip_addr_T *ip) {
+        int counter = 0;
+
+        if(ip == 0) {
+#ifndef MINGW
+                printf("%s","(null IP)\n");
+#else /* MINGW */
+                fprintf(LOG,"%s","(null IP)\n");
+#endif /* MINGW */
+                return;
+        }
+
+        if(ip->len == 4) {
+                for(counter = 0; counter < 3; counter++) {
+#ifndef MINGW
+                        printf("%d.",ip->ip[counter]);
+#else /* MINGW */
+                        fprintf(LOG,"%d.",ip->ip[counter]);
+#endif /* MINGW */
+                }
+#ifndef MINGW
+                printf("%d ",ip->ip[3]);
+#else /* MINGW */
+                fprintf(LOG,"%d ",ip->ip[3]);
+#endif /* MINGW */
+#ifdef IPV6
+        } else if(ip->len == 16) {
+                for(counter = 0; counter < 15; counter++) {
+#ifndef MINGW
+                        printf("%02x:",ip->ip[counter]);
+#else /* MINGW */
+                        fprintf(LOG,"%02x:",ip->ip[counter]);
+#endif /* MINGW */
+                }
+#ifndef MINGW
+                printf("%02x ",ip->ip[15]);
+#else /* MINGW */
+                fprintf(LOG,"%02x ",ip->ip[15]);
+#endif /* MINGW */
+#endif /* IPV6 */
+        } else {
+#ifndef MINGW
+                printf("%s%d","IP of length ",ip->len);
+#else /* MINGW */
+                fprintf(LOG,"%s%d","IP of length ",ip->len);
+#endif /* MINGW */
+        }
+}
+
+/* Log a char followed by an IP */
+void dw_log_ip(char *string, ip_addr_T *ip, int min_log_level) {
+        int32_t ll = key_n[DWM_N_verbose_level];
+        if(ll <= 0 || ll < min_log_level) {
+                return;
+        }
+
+#ifndef MINGW
+        printf("%s ",string);
+#else /* MINGW */
+        dw_win_time();
+        fprintf(LOG,"%s ",string);
+#endif /* MINGW */
+
+        dw_log_ip_p(ip);
+
+#ifndef MINGW
+        printf("%s","\n");
+#else /* MINGW */
+        fprintf(LOG,"%s","\n");
+#endif /* MINGW */
+
+}
+
 /* Log a string followed by the contents of a DwStr object */
 void dw_log_dwstr(char *s1, dw_str *s2, int min_log_level) {
         int32_t ll = key_n[DWM_N_verbose_level];
@@ -218,6 +292,21 @@ void dw_log_number(char *s1, int number, char *s2, int min_log_level) {
 
 }
 
+void dw_log_hex(char *s1, uint32_t number, int min_log_level) {
+        int32_t ll = key_n[DWM_N_verbose_level];
+
+        if(ll <= 0 || ll < min_log_level) {
+                return;
+        }
+
+#ifndef MINGW
+        printf("%s%x\n",s1,number);
+#else /* MINGW */
+        dw_win_time();
+        fprintf(LOG,"%s%x\n",s1,number);
+#endif /* MINGW */
+}
+
 /* Log 3 strings; input: Strings to log; minimum log level that we log these
  * strings at; this always logs and is run before Dwood2rc file is parsed */
 void dw_alog_3strings(char *s1, char *s2, char *s3) {
@@ -264,6 +353,7 @@ void dw_fatal(char *why) {
                 fprintf(LOG,"Fatal: Unknown fatal error\n");
 #endif /* MINGW */
         }
+        dw_log_close();
         exit(1);
 }
 
@@ -616,10 +706,16 @@ int32_t get_dns_qid(unsigned char *a, int len, int roy_arends_check) {
            /* a[8], a[9]: NSCOUNT */
            a[8] != 0 || a[9] != 0 ||
            /* a[10], a[11]: ARCOUNT */
-           a[10] != 0 || a[11] != 0))
+           a[10] != 0))
         {
                 return -1;
         }
+        if(roy_arends_check == 2 && a[11] == 1) {
+                return -2; /* Send NOTIMPL; OPT/EDNS not supported */
+        }
+        if(roy_arends_check == 2 && a[11] != 0) {
+                return -1; /* Drop packet */
+        }
         /* A DNS header is not considered kosher if it doesn't
          * have an question, nor any other RRs ; we look
          * at QDCOUNT, ANCOUNT, NSCOUNT, and ARCOUNT */
