@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2009 Sam Trenholme and others
+/* Copyright (c) 2007-2010 Sam Trenholme and others
  *
  * TERMS
  *
@@ -20,6 +20,7 @@
 #include "DwTcpSocket.h"
 #include "DwCompress.h"
 #include "DwDnsStr.h"
+#include "DwRecurse.h"
 
 /* Mararc parameters that are set in DwMararc.c */
 extern dw_str *key_s[];
@@ -55,10 +56,9 @@ extern int resurrections;
 extern int min_bind;
 extern int num_ports;
 extern int num_retries;
-extern int deliver_all;
 
 /* Other mararc parameters */
-dwd_dict *blacklist_dict;
+extern dwd_dict *blacklist_dict;
 
 #ifdef MINGW
 /* Needed for the Windows way of making a socket non-blocking */
@@ -176,14 +176,16 @@ SOCKET setup_server(sockaddr_all_T *server, ip_addr_T *addr) {
 /* Given a remote number, a C-string with the packet to send them, and
  * the length of that string, make a connection to a remote server */
 void make_remote_connection(int32_t n, unsigned char *packet, int len,
-     dw_str *query) {
+     dw_str *query, SOCKET x_socket_num) {
         sockaddr_all_T server;
         SOCKET s = 0;
-        int32_t rnum = -1;
-        ip_addr_T addr = {0,{0,0}};
+        int_fast32_t rnum = -1;
+        ip_addr_T addr = {0,{0,0},0,0};
+        int counter = 0;
         socklen_t inet_len = sizeof(struct sockaddr_in);
 
-        if(rem[n].socket != INVALID_SOCKET) { /* Already used (sanity check) */
+        if(rem[n].socket != x_socket_num || /* Already used (sanity check) */
+           rem[n].recurse_depth > 32) { /* Infinite recursion protection */
                 return;
         }
 
@@ -194,8 +196,21 @@ void make_remote_connection(int32_t n, unsigned char *packet, int len,
         }
 
         /* Get IP of remote server and open a socket to them */
-        addr = get_upstream_ip(query);
-        s = setup_server(&server,&addr);
+        for(counter = 0; counter < 7; counter++) { /* Don't stop on bad IP */
+                addr = get_upstream_ip(query,n);
+                if(addr.len > 0) {
+                        break;
+                }
+        }
+        dw_log_ip("Making connection to IP",&addr,128);
+        if(addr.len == 0) { /* Failed to get upstream IP */
+                return;
+        } else if(addr.len == 4 || addr.len == 16) { /* IPv4/IPv6 IP */
+                s = setup_server(&server,&addr);
+        } else if(addr.glueless != 0) { /* Glueless NS referral */
+                dwx_do_ns_glueless(addr,n);
+                return;
+        }
         if(s == INVALID_SOCKET) { /* Failed to make socket */
                 return;
         }
@@ -217,6 +232,9 @@ void make_remote_connection(int32_t n, unsigned char *packet, int len,
 
         /* OK, now note the pending connection */
         b_remote[n] = s;
+        if(rem[n].socket != INVALID_SOCKET) {
+                closesocket(rem[n].socket);
+        }
         rem[n].socket = s;
         rem[n].remote_id = rnum;
 }
@@ -260,6 +278,7 @@ int find_inflight_query(unsigned char *a, int len) {
         }
 
         query = dw_get_dname_type(a,12,len);
+        dwc_lower_case(query);
         if(query == 0) {
                 goto catch_find_inflight_query;
         }
@@ -307,9 +326,13 @@ void make_new_udp_connect(int b, unsigned char *a, int len, int num_alloc) {
                 return;
         }
         rem[b].query = dw_get_dname_type(a,12,len);
-        make_remote_connection(b,a,len,rem[b].query);
-
+        dwc_lower_case(rem[b].query);
         rem[b].local = malloc(num_alloc * sizeof(local_T *));
+        if(rem[b].local == 0) {
+                return;
+        }
+        make_remote_connection(b,a,len,rem[b].query,INVALID_SOCKET);
+
         for(counter = 0; counter < num_alloc; counter++) {
                 rem[b].local[counter] = 0;
         }
@@ -332,7 +355,7 @@ void make_new_udp_connect(int b, unsigned char *a, int len, int num_alloc) {
  * to be passed to the function. */
 int forward_local_udp_packet(SOCKET sock, int32_t local_id,
      ip_addr_T *from_ip, uint16_t from_port, unsigned char *a,
-     int len, int tcp_num) {
+     int len, int tcp_num, dw_str *orig_query) {
         int32_t b = 0;
         int num_alloc = 0;
 
@@ -363,12 +386,14 @@ int forward_local_udp_packet(SOCKET sock, int32_t local_id,
                 fflush(stdout);
 #endif /* INFLIGHT_VERBOSE */
         }
-        if(rem[b].socket == INVALID_SOCKET) {
+
+        rem[b].local[rem[b].num_locals - 1] = malloc(sizeof(local_T));
+        rem[b].local[rem[b].num_locals - 1]->orig_query = 0;
+        if(rem[b].socket == INVALID_SOCKET ||
+           rem[b].local[rem[b].num_locals - 1] == 0) {
                 reset_rem(b);
                 return -1;
         }
-
-        rem[b].local[rem[b].num_locals - 1] = malloc(sizeof(local_T));
         rem[b].local[rem[b].num_locals - 1]->from_socket = sock;
         rem[b].local[rem[b].num_locals - 1]->tcp_num = tcp_num;
         if(from_ip != 0) {
@@ -376,6 +401,9 @@ int forward_local_udp_packet(SOCKET sock, int32_t local_id,
         }
         rem[b].local[rem[b].num_locals - 1]->port = from_port;
         rem[b].local[rem[b].num_locals - 1]->local_id = local_id;
+        rem[b].local[rem[b].num_locals - 1]->glueless_type = 0;
+        rem[b].local[rem[b].num_locals - 1]->glueless_conn = -1;
+        rem[b].local[rem[b].num_locals - 1]->orig_query = dw_copy(orig_query);
         rem[b].die = get_time() + ((int64_t)timeout_seconds << 8);
         return 0;
 }
@@ -433,24 +461,37 @@ dw_str *make_dns_packet(dw_str *question, dw_str *answer, int32_t id) {
         dw_str *out = 0;
 
         is_nxdomain = dw_pop_u8(answer);
-        arcount = dw_pop_u16(answer);
-        nscount = dw_pop_u16(answer);
-        ancount = dw_pop_u16(answer);
+        if(is_nxdomain != TYPE_TRUNCATED &&
+                        is_nxdomain != TYPE_TRUNCATED_NXDOMAIN) {
+                arcount = dw_pop_u16(answer);
+                nscount = dw_pop_u16(answer);
+                ancount = dw_pop_u16(answer);
+        }
 
-        if(is_nxdomain == 0) {
+        if(is_nxdomain == 0 || is_nxdomain == 2) {
                 /* 0x8180: QR = 1; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
                  * Z = 0; RCODE = 0 */
                 out = make_dns_header(id,0x8180,ancount,nscount,arcount);
         } else if(is_nxdomain == 1) {
                 /* Same header as before, but with RCODE of "name error" */
                 out = make_dns_header(id,0x8183,ancount,nscount,arcount);
+        } else if(is_nxdomain == TYPE_TRUNCATED) {
+                /* Set TC to 1 */
+                out = make_dns_header(id,0x8380,0,0,0);
+        } else if(is_nxdomain == TYPE_TRUNCATED_NXDOMAIN) {
+                /* TC 1; RCODE "name error" */
+                out = make_dns_header(id,0x8383,0,0,0);
         } else {
                 goto catch_make_dns_packet;
         }
 
         if(dw_append(question,out) == -1 ||
-           dw_put_u16(out,1,-1) == -1 || /* QCLASS: 1 */
-           dw_append(answer,out) == -1) {
+           dw_put_u16(out,1,-1) == -1 /* QCLASS: 1 */ ) {
+                goto catch_make_dns_packet;
+        }
+        if(is_nxdomain != TYPE_TRUNCATED &&
+                        is_nxdomain != TYPE_TRUNCATED_NXDOMAIN &&
+                        dw_append(answer,out) == -1) {
                 goto catch_make_dns_packet;
         }
 
@@ -468,12 +509,13 @@ catch_make_dns_packet:
  * from the cache */
 int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
                          SOCKET sock, int32_t id, int resurrect,
-                         int tcp_num) {
+                         int tcp_num, dw_str *orig_query,
+                         ip_addr_T *from_ip) {
         dw_str *value = 0; /* Element in cache */
         dw_str *comp = 0; /* Compressed DNS packet */
         dw_str *packet = 0;
         socklen_t c_len = sizeof(struct sockaddr_in);
-        int ret = -1;
+        int ret = -1, type = 0, cache_type = 0;
 
 #ifdef IPV6
         if (client->Family == AF_INET6) {
@@ -481,9 +523,33 @@ int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
         }
 #endif
 
+        dw_log_dwstr("Looking in cache for query ",query,100);
+        type = dw_fetch_u16(query,-1);
+        if(type == RR_MX && key_n[DWM_N_reject_mx] != 0) {
+                dw_log_ip(
+        "Attempt to get MX record (possible spam zombie) from IP",from_ip,3);
+                return 1;
+        }
         dwc_process(cache,query,3); /* RR rotation, TTL aging, etc. */
         value = dwh_get(cache,query,resurrect,1);
-        comp = dwc_compress(query,value);
+        if(value == 0) {
+                dw_log_dwstr("Nothing found for ",query,100);
+                goto catch_get_reply_from_cache;
+        }
+        cache_type = dw_fetch_u8(value,-1);
+        if(cache_type != TYPE_TRUNCATED &&
+                        cache_type != TYPE_TRUNCATED_NXDOMAIN) {
+                comp = dwc_compress(query,value);
+        } else {
+                /* Immediately zap truncated from cache when fetched */
+                dwh_zap(cache,query,0,1);
+                if(client == 0) { /* DNS-over-TCP */
+                        ret = 2;
+                        goto catch_get_reply_from_cache;
+                }
+                comp = dw_copy(value);
+        }
+
         if(comp == 0) {
                 goto catch_get_reply_from_cache;
         }
@@ -495,7 +561,7 @@ int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
         }
 
         dw_log_dwstr_str("Fetching ",query," from cache",100);
-        packet = make_dns_packet(query,comp,id);
+        packet = make_dns_packet(orig_query,comp,id);
         if(packet == 0) {
                 goto catch_get_reply_from_cache;
         }
@@ -527,20 +593,21 @@ catch_get_reply_from_cache:
    expired element with the data we want. */
 void try_forward_local_udp_packet(SOCKET sock, int32_t local_id,
      ip_addr_T *from_ip, uint16_t from_port, unsigned char *packet,
-     int len, sockaddr_all_T *client,dw_str *query, int tcp_num) {
+     int len, sockaddr_all_T *client,dw_str *query, int tcp_num,
+     dw_str *orig_query) {
 
         /* If not cached, get a reply that we will cache and send back to
          * the client */
         if(forward_local_udp_packet(sock,local_id,from_ip,from_port,
-                                    packet,len,tcp_num) != -1) {
+                                    packet,len,tcp_num,orig_query) != -1) {
                 return; /* Success! */
         }
 
         /* OK, at this point it failed so we'll see if we get a
          * "resurrected" cache entry */
         if(resurrections == 1 &&
-           get_reply_from_cache(query,client,sock,local_id,1,tcp_num)
-             == 1) {
+           get_reply_from_cache(query, client, sock, local_id, 1, tcp_num,
+                        orig_query, from_ip) == 1) {
                 dw_log_string("Resurrected from cache",11);
                 return; /* Resurrected entry; we're done */
         }
@@ -556,10 +623,10 @@ void get_local_udp_packet(SOCKET sock) {
         int len = 0;
         sockaddr_all_T client;
         socklen_t c_len = 0;
-        ip_addr_T from_ip = {0,{0,0}};
+        ip_addr_T from_ip = {0,{0,0},0,0};
         uint16_t from_port = 0;
         int32_t local_id = -1;
-        dw_str *query = 0;
+        dw_str *query = 0, *orig_query = 0;
 #ifdef VALGRIND_NOERRORS
         memset(packet,0,522);
 #endif /* VALGRIND_NOERRORS */
@@ -569,6 +636,10 @@ void get_local_udp_packet(SOCKET sock) {
         len = recvfrom(sock,(void *)packet,520,0,(struct sockaddr *)&client,
                        &c_len);
 
+        if(len < 12) {
+                goto catch_get_local_udp_packet;
+        }
+
         from_port = get_from_ip_port(&from_ip,&client);
 
         if(check_ip_acl(&from_ip) != 1) { /* Drop unauthorized packets */
@@ -576,6 +647,24 @@ void get_local_udp_packet(SOCKET sock) {
         }
 
         local_id = get_dns_qid(packet, len, 2);
+        if(local_id == -2) { /* Immediate NOTIMPL for EDNS requests w/ OPT */
+#ifdef STRICT_RFC2671_COMPLIANCE
+                packet[2] = 0x81; /* QR = 1; Op = 0; AA = 0; TC = 0; RD = 1 */
+                packet[3] = 0x84; /* RA = 1; Z = 0; RCODE = "notimpl" (4) */
+                sendto(sock,(void *)packet,len,0,(struct sockaddr *)&client,
+                                c_len);
+                goto catch_get_local_udp_packet;
+#else /* STRICT_RFC2671_COMPLIANCE */
+                query = dw_get_dname_type(packet,12,len);
+                if(query == 0) {
+                        goto catch_get_local_udp_packet;
+                }
+                len = query->len + 14;
+                packet[11] = 0; /* We no longer have final question */
+                dw_destroy(query);
+                local_id = get_dns_qid(packet, len, 2);
+#endif /* STRICT_RFC2671_COMPLIANCE */
+        }
         if(local_id < 0 || len < 13) { /* Invalid remote packet */
                 goto catch_get_local_udp_packet;
         }
@@ -583,27 +672,34 @@ void get_local_udp_packet(SOCKET sock) {
         /* See if we have something in the cache; destroyed at end of
          * function */
         query = dw_get_dname_type(packet,12,len);
+        orig_query = dw_copy(query);
+        dwc_lower_case(query);
         dw_log_dwstr("Got DNS query for ",query,100); /* Log it */
         if(query == 0) {
                 goto catch_get_local_udp_packet;
         }
-        if(get_reply_from_cache(query,&client,sock,local_id,0,-1)
-                == 1) {
+        if(get_reply_from_cache(query, &client, sock, local_id, 0, -1,
+                        orig_query, &from_ip) == 1) {
                 goto catch_get_local_udp_packet; /* In cache; we're done */
         }
 
         /* Nothing in cache; lets try sending the packet upstream */
         try_forward_local_udp_packet(sock,local_id,&from_ip,from_port,
-                        packet,len,&client,query,INVALID_SOCKET);
+                        packet,len,&client,query,INVALID_SOCKET,
+                        orig_query);
 
 catch_get_local_udp_packet:
         if(query != 0) {
                 dw_destroy(query);
                 query = 0;
         }
+        if(orig_query != 0) {
+                dw_destroy(orig_query);
+                query = 0;
+        }
 }
 
-/* Forward a remote reply back to the client */
+/* Forward a remote reply back to the client over UDP */
 void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip,
                 int local_num) {
         sockaddr_all_T client;
@@ -612,7 +708,8 @@ void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip,
         if(r_ip == 0) {
                 return;
         }
-        if(r_ip->local[local_num]->from_socket == INVALID_SOCKET) {
+        if(r_ip->local[local_num]->from_socket == INVALID_SOCKET ||
+                        r_ip->local[local_num]->glueless_type != 0) {
                 return;
         }
         memset(&client,0,sizeof(client));
@@ -646,44 +743,67 @@ void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip,
 
 /* Add a reply we have received from the remote (upstream) DNS server to
  * the cache */
-int cache_dns_reply(unsigned char *packet, int count) {
+int cache_dns_reply(unsigned char *packet, int count, int b, int truncated) {
         int32_t ttl = 60;
         int32_t ancount = 0;
         int is_nxdomain = 0;
         dw_str *question = 0, *answer = 0;
         dw_str *decomp = 0;
-        int ret = -1;
+        int ret = -1, qtype = -1;
 
         question = dw_get_dname_type(packet,12,count);
-        dw_log_dwstr("Caching reply for ",question,100);
-        if((packet[3] & 0x0f) == 3) {
+        dwc_lower_case(question);
+        qtype = dw_fetch_u16(question,-1);
+        dw_log_dwstr("Caching a reply for query ",question,1000);
+        if((packet[3] & 0x0f) == 2) { /* Server FAIL */
+                ret = -1; /* Bad return value; do not cache */
+                goto catch_cache_dns_reply;
+        } else if((packet[3] & 0x0f) == 3) { /* Name error/NXDOMAIN */
                 is_nxdomain = 1;
         }
-        answer = dw_packet_to_cache(packet,count,is_nxdomain);
-        decomp = dwc_decompress(question,answer);
-        if(blacklist_dict != 0 && has_bad_ip(decomp,blacklist_dict)) {
-                ret = -2; /* Tell caller we need synth "not there" */
-                goto catch_cache_dns_reply;
-        }
-        ancount = dw_cachepacket_to_ancount(answer);
-        if(ancount == 0) {
-                ancount = 32; /* So we can correctly cache negative answers */
-        }
+        if(truncated == 1) {
+                is_nxdomain += 3;
+                answer = dw_create(2);
+                if(dw_put_u8(answer, is_nxdomain, 0) == -1) {
+                        goto catch_cache_dns_reply;
+                }
+                dwh_add(cache,question,answer,7,1);
+                ret = 1;
+        } else {
+                answer = dw_packet_to_cache(packet,count,is_nxdomain);
+                decomp = dwc_decompress(question,answer);
+                if(dwc_has_bad_ip(decomp,blacklist_dict) ||
+                   (qtype == 28 /* AAAA */ && key_n[DWM_N_reject_aaaa] == 1)) {
+                        ret = -2; /* Tell caller we need synth "not there" */
+                        goto catch_cache_dns_reply;
+                }
+                ancount = dw_cachepacket_to_ancount(answer);
+                if(ancount == 0) {
+                        ancount = 32; /* Correct negative answer caching */
+                }
 
-        if(question == 0 || answer == 0 || ancount == -1) {
-                goto catch_cache_dns_reply;
+                if(question == 0 || answer == 0 || ancount == -1) {
+                        goto catch_cache_dns_reply;
+                }
+
+                ttl = dw_get_a_dnsttl(answer,0,31536000,ancount);
+                if(ttl == -1 && decomp != 0 && decomp->len == 7) {
+                        ttl = 30; /* Special case: Blank reply upstream */
+                }
+                if(ttl == -1) {
+                        goto catch_cache_dns_reply;
+                }
+                if(ttl < 30) {
+                        ttl = 30;
+                }
+
+                /* Routines in DwRecurse.c process the packet and let us know
+                 * what kind of packet we got upstream (so we know how to
+                 * continue)
+                 */
+                ret = dwx_cache_reply(cache,question,decomp,ttl,b);
         }
 
-        ttl = dw_get_a_dnsttl(answer,0,31536000,ancount);
-
-        if(ttl == -1) {
-                goto catch_cache_dns_reply;
-        }
-
-        /* Physically put the data in the cache */
-        dwh_add(cache,question,decomp,ttl,1);
-        ret = 1;
-
 catch_cache_dns_reply:
         if(question != 0) {
                 dw_destroy(question);
@@ -714,7 +834,7 @@ int verify_dns_packet(int b, unsigned char *packet, int len) {
                 goto catch_verify_dns_packet;
         }
 
-        /* Should we make this case-insensitive? Probably not. */
+        dwc_lower_case(question); /* Case-insensitive comparison */
         if(dw_issame(question,rem[b].query) != 1) {
                 goto catch_verify_dns_packet;
         }
@@ -730,7 +850,8 @@ catch_verify_dns_packet:
 }
 
 /* Make the actual answer for a synthetic "not there" reply */
-unsigned char *make_synth_not_there_answer(unsigned char *a, int count) {
+unsigned char *make_synth_not_there_answer(unsigned char *a, int count,
+                int type) {
         /* This is the answer for a "not there" reply */
         unsigned char not_there[41] =
         "\xc0\x0c" /* Name */
@@ -745,6 +866,17 @@ unsigned char *make_synth_not_there_answer(unsigned char *a, int count) {
         int counter = 0;
 
         answer = malloc(count + 43);
+        if(answer == 0) {
+                return 0;
+        }
+
+        if(type == 1) { /* Special case: Return just synth "not there" */
+                for(counter = 0; counter < 40; counter++) {
+                        answer[counter] = not_there[counter];
+                }
+                return answer;
+        }
+
         /* Copy the header they sent us to our reply */
         for(counter = 0; counter < 12 && counter < count; counter++) {
                 answer[counter] = a[counter];
@@ -789,7 +921,7 @@ void make_synth_not_there(int b, SOCKET sock, unsigned char *a, int count) {
         }
 
         /* Copy original header and question in to answer */
-        answer = make_synth_not_there_answer(a,count);
+        answer = make_synth_not_there_answer(a,count,0);
         if(answer == 0) {
                 return;
         }
@@ -803,6 +935,9 @@ void make_synth_not_there(int b, SOCKET sock, unsigned char *a, int count) {
 
         /* Send the reply(s) */
         for(local_num = 0; local_num < rem[b].num_locals; local_num++) {
+                if(rem[b].local[local_num]->glueless_type != 0) {
+                        continue;
+                }
                 /* Copy ID over */
                 answer[0] = (rem[b].local[local_num]->local_id >> 8) & 0xff;
                 answer[1] = (rem[b].local[local_num]->local_id) & 0xff;
@@ -818,13 +953,71 @@ void make_synth_not_there(int b, SOCKET sock, unsigned char *a, int count) {
 
         /* Reset the pending remote connection */
         closesocket(b_remote[b]);
-        b_remote[b] = -1;
+        b_remote[b] = INVALID_SOCKET;
         reset_rem(b);
 
         /* Free allocated memory */
         free(answer);
 }
 
+/* Given a query, a socket, a remote connection_number (b), and a
+ * local connection number for that remote connection (l), we get
+ * a reply from the cache to send to the client in question.
+ *
+ * Don't confuse this with get_reply_from_cache, which does the same
+ * thing with different arguments; this is a wrapper for
+ * get_reply_from_cache() which is called from get_rem_udp_packet_core()
+ */
+int send_reply_from_cache(unsigned char *a, ssize_t count, int b, int l) {
+        dw_str *query = 0;
+        sockaddr_all_T client;
+        int conn_num = 0;
+        int ret = 1;
+
+        query = dw_get_dname_type(a,12,count);
+        if(query == 0 || rem[b].local == 0 || rem[b].local[l] == 0) {
+                goto catch_send_reply_from_cache;
+        }
+        memset(&client,0,sizeof(client));
+        if (rem[b].local[l]->glueless_type != 0) {
+                conn_num = rem[b].local[l]->glueless_conn;
+                if(rem[b].local[l]->glueless_type == 1) { /* Glueless NS */
+                        if(rem[conn_num].child_id == b) { /* Sanity check */
+                                dwx_glueless_done(query, conn_num);
+                        }
+                        ret = 4; /* Yes, we want to end the connection */
+                } else if(rem[b].local[l]->glueless_type == 2) { /* CNAME */
+                        dwx_incomplete_cname_done(query, b, l);
+                        ret = 4; /* End connection */
+                }
+                goto catch_send_reply_from_cache;
+        } else if (rem[b].local[l]->ip.len == 4) {
+                client.V4.sin_family = AF_INET;
+                client.V4.sin_port = htons(rem[b].local[l]->port);
+                memcpy(&client.V4.sin_addr, rem[b].local[l]->ip.ip, 4);
+#ifdef IPV6
+        } else if(rem[b].local[l]->ip.len == 16) {
+                client.V6.sin6_family = AF_INET6;
+                client.V6.sin6_port = htons(rem[b].local[l]->port);
+                memcpy(&client.V6.sin6_addr, rem[b].local[l]->ip.ip,
+                        16);
+#endif
+        } else {
+                ret = 1;
+                goto catch_send_reply_from_cache;
+        }
+
+        get_reply_from_cache(query,&client,rem[b].local[l]->from_socket,
+                        rem[b].local[l]->local_id, 0,-1,
+                        rem[b].local[l]->orig_query, 0);
+
+catch_send_reply_from_cache:
+        if(query != 0) {
+                dw_destroy(query);
+        }
+        return ret;
+}
+
 /* Core part of code that gets and processes remote DNS requests */
 
 int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
@@ -836,34 +1029,46 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
                 return -1; /* Not a DNS packet at all */
         }
 
+        if(rem[b].local[l] == 0) {
+                return -1;
+        }
+
         if((a[2] & 0x02) == 0x00) { /* If not truncated */
                 fflush(stdout);
-                cache_dns_reply_return_value = cache_dns_reply(a,count);
+                cache_dns_reply_return_value = cache_dns_reply(a,count,b,0);
                 if(cache_dns_reply_return_value == -2) { /* Make synth NX */
                         make_synth_not_there(b,sock,a,count);
                         return -1; /* Bad reply and they got a Synth NX */
                 }
-                if(cache_dns_reply_return_value < 0 && deliver_all == 0) {
+                if(cache_dns_reply_return_value < 0) {
                         return -1; /* Bad reply */
                 }
+                if(cache_dns_reply_return_value >= 16) { /* Don't forward */
+                        return 3; /* Don't forward to end user; cache
+                                   * and next step */
+                }
         } else if(rem[b].local[l]->tcp_num != -1 &&
                         key_n[DWM_N_tcp_listen] == 1) {
                 /* Send a DNS-over-TCP packet to handle a truncated reply */
                 tcp_truncated_retry(rem[b].local[l]->tcp_num, rem[b].query,
-                        rem[b].local[l]->local_id);
+                        rem[b].local[l]->local_id,b,rem[b].is_upstream);
                 /* Give the UDP connection more time before timeout, so
                  * we can fully process the TCP connection */
                 rem[b].die = get_time() + ((int64_t)timeout_seconds << 10);
                 return 2; /* Don't kill pending UDP connection */
+        } else { /* Truncated over UDP; just given them a blank "truncated"
+                  * reply */
+                cache_dns_reply(a,count,b,1);
         }
 
         /* Now make sure the ID is the same as the one the client
          * originally sent us */
         set_dns_qid(a,count,rem[b].local[l]->local_id);
 
-        /* Send the answer we got to the appropriate local connection */
-        if(rem[b].local[l]->tcp_num == -1 ) {
-                forward_remote_reply(a,count,&rem[b], l);
+        /* Send the answer we just cached over appropriate local connection */
+        if(rem[b].local[l]->tcp_num == -1 ||
+                        rem[b].local[l]->glueless_type != 0) {
+                return send_reply_from_cache(a,count,b,l);
         } else if(key_n[DWM_N_tcp_listen] == 1) {
                 tcp_return_reply(rem[b].local[l]->tcp_num,(void *)a,count);
         }
@@ -871,7 +1076,7 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
         return 1;
 }
 
-/* Get and process a remote DNS request */
+/* Get and process a remote DNS packet (one sent upstream to us) */
 void get_remote_udp_packet(int b, SOCKET sock) {
         ssize_t count;
         unsigned char a[520];
@@ -896,7 +1101,7 @@ void get_remote_udp_packet(int b, SOCKET sock) {
                 core_ret = get_rem_udp_packet_core(a,count,b,sock,l);
                 if(core_ret == -1) {
                         return;
-                } else if(core_ret == 2) {
+                } else if(core_ret == 2 || core_ret == 3) {
                         kill = 0;
                 }
         }
@@ -904,7 +1109,7 @@ void get_remote_udp_packet(int b, SOCKET sock) {
         /* Reset this pending remote connection if needed */
         if(kill == 1) {
                 closesocket(b_remote[b]);
-                b_remote[b] = -1;
+                b_remote[b] = INVALID_SOCKET;
                 reset_rem(b);
         }
 }
@@ -914,6 +1119,11 @@ void get_remote_udp_packet(int b, SOCKET sock) {
 void server_fail_noreply(int a, int local_num) {
         dw_str *packet = 0;
 
+        if(rem[a].local[local_num] == 0 ||
+           rem[a].local[local_num]->glueless_type != 0) {
+                return;
+        }
+
         /* 0x8182: QR = 1; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
          *         Z = 0; RCODE = "server fail" (2) */
         packet = make_dns_header(rem[a].local[local_num]->local_id,0x8182,
@@ -926,7 +1136,8 @@ void server_fail_noreply(int a, int local_num) {
                 goto catch_server_fail_noreply;
         }
 
-        if(rem[a].local[local_num]->tcp_num == -1) {
+        if(rem[a].local[local_num]->tcp_num == -1 ||
+                        rem[a].local[local_num]->glueless_type != 0) {
                 forward_remote_reply((unsigned char *)packet->str,
                         packet->len,&rem[a],0);
         } else if(key_n[DWM_N_tcp_listen] == 1) {
