@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2006 Sam Trenholme
+/* Copyright (c) 2002-2007 Sam Trenholme
  *
  * TERMS
  *
@@ -48,6 +48,7 @@
 
 extern int log_level;
 extern int no_cname_warnings;
+extern int dns_records_served;
 int cname_warnings_shown = 0;
 
 
@@ -716,7 +717,7 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
     /* Pass 1: fill up the hash with the AN and NS answers */
 
     /* Get all of the keys in the csv1 hash */
-    csvs = (mhash *)dvar_raw(dq_keyword2n("csv1")); 
+    csvs = (mhash *)dvar_raw(dq_keyword2n("csv1"));
 
     if(csvs == 0) {
         /* A csv1 hash is no longer mandatory, now that MaraDNS has recursive
@@ -804,7 +805,7 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
                 /* Get the line */
                 result = js_buf_getline(desc,line);
                 if(result == JS_ERROR) {
-                    js_destroy(udpzone); js_destroy(filename); 
+                    js_destroy(udpzone); js_destroy(filename);
                     js_destroy(line);
                     js_destroy(pline); js_destroy(soaline); js_destroy(query);
                     js_destroy(data); js_destroy(zone); js_dealloc(desc);
@@ -815,7 +816,7 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
                     show_esc_stdout(filename);
                     printf("\n");
                     /* "line number ... is too long in file " */
-                    js_destroy(udpzone); js_destroy(filename); 
+                    js_destroy(udpzone); js_destroy(filename);
                     js_destroy(line);
                     js_destroy(pline); js_destroy(soaline); js_destroy(query);
                     js_destroy(data); js_destroy(zone); js_dealloc(desc);
@@ -836,9 +837,9 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
                        only having SOA and NS records */
                     if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA) {
 
-                        js_destroy(udpzone); js_destroy(filename); 
+                        js_destroy(udpzone); js_destroy(filename);
                         js_destroy(line);
-                        js_destroy(pline); js_destroy(soaline); 
+                        js_destroy(pline); js_destroy(soaline);
                         js_destroy(query);
                         js_destroy(data); js_destroy(zone); js_dealloc(desc);
                         return JS_ERROR;
@@ -853,7 +854,7 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
 
                 /* Process the % character and any \ sequences */
                 if(bs_process(line,pline,zone) == JS_ERROR) {
-                    js_destroy(udpzone); js_destroy(filename); 
+                    js_destroy(udpzone); js_destroy(filename);
                     js_destroy(line);
                     js_destroy(pline); js_destroy(soaline); js_destroy(query);
                     js_destroy(data); js_destroy(zone); js_dealloc(desc);
@@ -903,11 +904,11 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
                        section of the answer) */
                     else if(rrnum == 1) {
                         if(js_copy(pline,soaline) == JS_ERROR) {
-                            js_destroy(udpzone); js_destroy(filename); 
+                            js_destroy(udpzone); js_destroy(filename);
                             js_destroy(line);
-                            js_destroy(pline); js_destroy(soaline); 
+                            js_destroy(pline); js_destroy(soaline);
                             js_destroy(query);
-                            js_destroy(data); js_destroy(zone); 
+                            js_destroy(data); js_destroy(zone);
                             js_dealloc(desc);
                             return JS_ERROR;
                             }
@@ -927,13 +928,13 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
                     else if((rr_type != RR_NS || !js_issame(query,udpzone))
                             && in_ns) {
                         /* Add the SOA after the zone's NS records */
-                        if(parse_csv1_line(soaline,query,data,&ttl) != 
-			   RR_SOA) {
-                            js_destroy(udpzone); js_destroy(filename); 
+                        if(parse_csv1_line(soaline,query,data,&ttl) !=
+                           RR_SOA) {
+                            js_destroy(udpzone); js_destroy(filename);
                             js_destroy(line);
-                            js_destroy(pline); js_destroy(soaline); 
+                            js_destroy(pline); js_destroy(soaline);
                             js_destroy(query);
-                            js_destroy(data); js_destroy(zone); 
+                            js_destroy(data); js_destroy(zone);
                             js_dealloc(desc);
                             return JS_ERROR;
                             }
@@ -974,15 +975,15 @@ int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
     return JS_SUCCESS;
 }
 
-/* This program parses all of the zone files and adds the records 
+/* This program parses all of the zone files and adds the records
  * to the MaraDNS database */
 int populate_main(mhash *maintable, js_string *error, int recursive) {
-	/* Pass 1: parse all of the zone files */
-	parse_csv1s(maintable,error,recursive);
-	csv2_parse_main_bighash(maintable,0);
-	
-	/* Pass 2: add the appropriate A records */
-	return add_an(maintable,error);
+        /* Pass 1: parse all of the zone files */
+        parse_csv1s(maintable,error,recursive);
+        csv2_parse_main_bighash(maintable,0);
+
+        /* Pass 2: add the appropriate A records */
+        return add_an(maintable,error);
 }
 
 /* Warn about the existance of a DDIP record
@@ -1076,7 +1077,9 @@ int make_ddip(mhash *bighash, js_string *query) {
    AR section (AN section with CNAMEs)
    Input: A pointer to the hash to populate, a pointer to the string to
           put an error message in
-   Output: JS_ERROR on error, -2 on non-fatal error, JS_SUCCESS on success
+   Output: JS_ERROR on error, -2 on non-fatal error, JS_SUCCESS when there
+           is one or more authoritative records, 0 when there are no
+           authoritative records
  */
 
 int add_an(mhash *bighash, js_string *error) {
@@ -1137,12 +1140,12 @@ int add_an(mhash *bighash, js_string *error) {
                         }
                     } while ((record = record->next) != 0);
                 }
-	    /* If it is a CNAME query, proceed to find the assosciated
-	     * ANY value */
+            /* If it is a CNAME query, proceed to find the assosciated
+             * ANY value */
             } else if(qtype == RR_CNAME) {
-		/* Look for a corresponding ANY value (which needs to be 
-		 * there) */
-		record = qdata.value;
+                /* Look for a corresponding ANY value (which needs to be
+                 * there) */
+                record = qdata.value;
                 do {
                     if(answer_ip_query(qtype,record->data,a_query) == JS_ERROR)
                         continue;
@@ -1156,21 +1159,21 @@ int add_an(mhash *bighash, js_string *error) {
                         if(record != 0 && adata.value != 0)
                             record->list = adata.value;
                         }
-		    else if(no_cname_warnings == 0 && 
-			       cname_warnings_shown < 10) {
-			printf("Warning: The CNAME record ");
-			human_readable_dns_query(query,1);
-			printf(" is a dangling CNAME record.\n"
+                    else if(no_cname_warnings == 0 &&
+                               cname_warnings_shown < 10) {
+                        printf("Warning: The CNAME record ");
+                        human_readable_dns_query(query,1);
+                        printf(" is a dangling CNAME record.\n"
 "Please read the FAQ entry about FAQ entry about dangling CNAME records.\n");
-			cname_warnings_shown++;
-		        }
-		    else if(no_cname_warnings == 0 &&
-			      cname_warnings_shown == 10) {
-		       printf("Warning: More dangling CNAME records exist.\n");
+                        cname_warnings_shown++;
+                        }
+                    else if(no_cname_warnings == 0 &&
+                              cname_warnings_shown == 10) {
+                       printf("Warning: More dangling CNAME records exist.\n");
                        cname_warnings_shown++;
                        }
                     } while ((record = record->next) != 0);
-	        }
+                }
         } while(mhash_nextkey(bighash,query) != 0);
     js_destroy(a_query);
     js_destroy(query);
@@ -1255,10 +1258,10 @@ int answer_ip_query(int qtype, js_string *qu, js_string *ar) {
              data is there, do nothing.
 
           The "perms" (what set of IPs are allowed to view this record) of
-	  the record in question; if a given perm bit is one, only the IP
-	  corresponding to the perm bit in question is allowed to view this
-	  record.  If it is zero, this is the default value: any IP on
-	  the internet can see the record in question.
+          the record in question; if a given perm bit is one, only the IP
+          corresponding to the perm bit in question is allowed to view this
+          record.  If it is zero, this is the default value: any IP on
+          the internet can see the record in question.
 
    Output: JS_ERROR on error, JS_SUCCESS on success
 */
@@ -1284,12 +1287,15 @@ int add_rr_to_bighash(mhash *bighash,
 
     /* We are not allowed to add RR_ANY data to the bighash */
     if(qtype == RR_ANY)
-	return JS_ERROR;
+        return JS_ERROR;
 
     /* Nor are we allowed to add "AXFR" nor "IXFR" data to the bighash */
-    if(qtype == 251 /* IXFR */ || qtype == 252 /* AXFR */) 
+    if(qtype == 251 /* IXFR */ || qtype == 252 /* AXFR */)
         return JS_ERROR;
 
+    /* Increment the number of DNS records we are serving */
+    dns_records_served++;
+
     /* See if the data is authoritative or not */
     /* Change the type so the two strings are the same */
     if(change_rtype(zone,qtype) == JS_ERROR) {
@@ -1336,8 +1342,8 @@ int add_rr_to_bighash(mhash *bighash,
        the first NS record for this zone if this is a non-NS record.  */
     if(hash_spot.value == 0) {
         /* Add the data */
-        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms) 
-			== JS_ERROR)
+        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         if(mhash_add_ns(bighash,query,zone) == JS_ERROR)
             return JS_ERROR;
@@ -1358,8 +1364,8 @@ int add_rr_to_bighash(mhash *bighash,
        the beginning in the case of NS records) */
     if((point->authoritative == 1 && has_authority == 1) ||
        (point->authoritative == 0 && has_authority == 0)) {
-        if(mhash_add_rr(bighash,query,data,ttl,has_authority,0,perms) 
-			== JS_ERROR)
+        if(mhash_add_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         return JS_SUCCESS;
         }
@@ -1375,18 +1381,18 @@ int add_rr_to_bighash(mhash *bighash,
         nukeme = (rr *)mhash_undef(bighash,query);
         if(nukeme == 0)
             return JS_ERROR;
-	/* Remove this entry from the corresponding
-	 * ANY chain */
-	if(any_zap_rr(bighash,query,nukeme) == JS_ERROR) {
-	    return JS_ERROR;
-	    }
-	/* Remove the entry once and for all */
+        /* Remove this entry from the corresponding
+         * ANY chain */
+        if(any_zap_rr(bighash,query,nukeme) == JS_ERROR) {
+            return JS_ERROR;
+            }
+        /* Remove the entry once and for all */
         if(js_dealloc(nukeme) == JS_ERROR)
             return JS_ERROR;
 
         /* Add the data */
-        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms) 
-			== JS_ERROR)
+        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         if(mhash_add_ns(bighash,query,zone) == JS_ERROR)
             return JS_ERROR;
